# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

"""Settings used in the BlockerBugs application"""

import os
import sys


class Config(object):
    """Default configuration for the whole application. Default values are tailored more towards a
    development environment.
    """
    PRODUCTION = False
    DEBUG = True
    SECRET_KEY = None  # REPLACE THIS WITH A RANDOM STRING
    SQLALCHEMY_DATABASE_URI = 'sqlite://'  # just 'sqlite://' without a path means an in-memory db
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    BUGZILLA_URL = 'https://bugzilla.stage.redhat.com/'
    BUGZILLA_XMLRPC = BUGZILLA_URL + 'xmlrpc.cgi'
    KOJI_URL = 'http://koji.stg.fedoraproject.org/'
    BODHI_URL = 'https://bodhi.stg.fedoraproject.org/'
    FAS_ENABLED = False
    """When FAS is not enabled, a fake stub is used instead, which allows the user to log in (under
    the credentials+groups defined here in this config) without authentication."""
    FAS_ADMIN_GROUP = 'qa-admin'
    FAS_USER = ''
    FAS_PASSWORD = ''
    FAS_HTTPS_REQUIRED = False
    FAS_CHECK_CERT = False
    FAS_BASE_URL = 'https://admin.stg.fedoraproject.org/accounts/'
    LOGFILE = '/var/log/blockerbugs/blockerbugs.log'
    FILE_LOGGING = False
    SYSLOG_LOGGING = False
    STREAM_LOGGING = True
    SHOW_DB_URI = False
    BEHIND_PROXY = False
    FEDMENU_URL = ""
    FEDMENU_DATA_URL = ""
    BLOCKERBUGS_URL = "https://qa.fedoraproject.org/blockerbugs/"
    BLOCKERBUGS_API = "{}api/v0/".format(BLOCKERBUGS_URL)
    PAGURE_URL = "https://stg.pagure.io/"
    PAGURE_API = "https://stg.pagure.io/api/0/"
    PAGURE_REPO = "fedoraqa-test"
    PAGURE_REPO_TOKEN = "YOUR SECRET API TOKEN FROM PROJECT SETTINGS"
    PAGURE_REPO_WEBHOOK_KEY = "YOUR WEBHOOK KEY PROJECT SETTINGS"
    PAGURE_BOT_USERNAME = 'blockerbot'
    PAGURE_BOT_ENABLED = True
    PAGURE_BOT_LOOP_THRESHOLD = 2
    PAGURE_DISCUSSION_TITLE = "[$component] $summary | rhbz#$bugid"
    PAGURE_DISCUSSION_CONTENT = '''\
Bug details: ** $bug_url **
Information from [BlockerBugs App]({}):
![$bugid]($bug_img)
#### Current vote summary
$vote_summary

To learn how to vote, see:
$pagure_url$pagure_repo
A quick example: `BetaBlocker +1` (where the tracker name is one of \
`BetaBlocker`/`FinalBlocker`/`BetaFE`/`FinalFE`/`0Day`/`PreviousRelease` \
and the vote is one of `+1`/`0`/`-1`)
'''.format(BLOCKERBUGS_URL)


class ProductionConfig(Config):
    """A `Config` subclass intended for a production deployment"""
    PRODUCTION = True
    DEBUG = False
    BUGZILLA_URL = 'https://bugzilla.redhat.com/'
    BUGZILLA_XMLRPC = BUGZILLA_URL + 'xmlrpc.cgi'
    KOJI_URL = 'http://koji.fedoraproject.org/'
    BODHI_URL = 'https://bodhi.fedoraproject.org/'
    FAS_ENABLED = True
    FAS_HTTPS_REQUIRED = True
    FAS_BASE_URL = 'https://admin.fedoraproject.org/accounts/'
    PAGURE_URL = "https://pagure.io/"
    PAGURE_API = "https://pagure.io/api/0/"
    PAGURE_REPO = "fedora-qa/blocker-review"
    SHOW_DB_URI = False


class DevelopmentConfig(Config):
    """A `Config` subclass intended for a development environment"""
    TRAP_BAD_REQUEST_ERRORS = True
    SQLALCHEMY_DATABASE_URI = 'sqlite:////var/tmp/blockerbugs_db.sqlite'
    SHOW_DB_URI = True


class TestingConfig(Config):
    """A `Config` subclass intended for running the test suite"""
    TESTING = True
    SHOW_DB_URI = True


def openshift_config(config_object, openshift_production):
    # First, get db details from env
    try:
        config_object["SQLALCHEMY_DATABASE_URI"] = "postgresql+psycopg2://%s:%s@%s:%s/%s" % (
            os.environ["POSTGRESQL_USER"],
            os.environ["POSTGRESQL_PASSWORD"],
            os.environ["POSTGRESQL_SERVICE_HOST"],
            os.environ["POSTGRESQL_SERVICE_PORT"],
            os.environ["POSTGRESQL_DATABASE"]
        )
        config_object["SECRET_KEY"] = os.environ["SECRET_KEY"]
    except(KeyError):
        print("OpenShift mode enabled but required values couldn't be fetched. "
              "Check, if you have these variables defined in you env: "
              "(POSTGRESQL_[USER, PASSWORD, DATABASE, SERVICE_HOST, SERVICE_PORT], SECRET_KEY)", file=sys.stderr)
        sys.exit(1)

    # And then try to get more data from OpenShift env
    try:
        config_object["FAS_PASSWORD"] = os.environ["FAS_PASSWORD"]
        config_object["FAS_USER"] = os.environ["FAS_USER"]
        config_object["FAS_ADMIN_GROUP"] = os.environ["FAS_ADMIN_GROUP"]
        config_object["PAGURE_REPO_TOKEN"] = os.environ["PAGURE_REPO_TOKEN"]
        config_object["PAGURE_REPO_WEBHOOK_KEY"] = os.environ["PAGURE_REPO_WEBHOOK_KEY"]
        config_object["PAGURE_BOT_USERNAME"] = os.environ["PAGURE_BOT_USERNAME"]
    except(KeyError):
        print("Expected aditional data to be defined OpenShift env: "
              "FAS_PASSWORD, FAS_USER, FAS_ADMIN_GROUP, OPENSHIFT_PROD.", file=sys.stderr)
        # In production OpenShift, exit on missing additional details
        if openshift_production == "1":
            sys.exit(1)

    # Get some more data from OpenShift, if set
    if os.getenv("PAGURE_BOT_ENABLED") == 0:
        config_object["PAGURE_BOT_ENABLED"] = False
    if os.getenv("PAGURE_URL"):
        config_object["PAGURE_URL"] = os.getenv("PAGURE_URL")
    if os.getenv("PAGURE_API"):
        config_object["PAGURE_API"] = os.getenv("PAGURE_API")
