# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

"""ORM class for bug information stored in the database"""

import datetime
import json
from typing import Any, Optional

from blockerbugs import db, BaseModel, models
import blockerbugs.models.milestone as model_milestone
import blockerbugs.models.update as model_update


class Bug(BaseModel):
    """A representation of a Bugzilla ticket, with some additional attributes.

    Note: While a particular bug (with a given `bugid`) can exist only once in Bugzilla, a matching
    `Bug` instance is not unique in BlockerBugs. It is only unique in relation to a certain
    `Milestone`. So e.g. there can only be a single `Bug` with `bugid=123456` under F35-Beta, but
    there can be another `Bug` with `bugid=123456` under F35-Final, or F36-Beta.
    """

    id = db.Column(db.Integer, primary_key=True)
    bugid = db.Column(db.Integer, unique=False)
    """The bug number, e.g. 123456"""
    url = db.Column(db.String(1024), unique=False)
    summary = db.Column(db.String(2048), unique=False)
    status = db.Column(db.String(80), unique=False)
    """E.g. 'NEW' or 'ASSIGNED'"""
    component = db.Column(db.String(80), unique=False)
    active = db.Column(db.Boolean(create_constraint=True, name='active_bool'))
    """An alias for 'open', i.e. `True` when open, `False` when closed"""
    last_bug_sync = db.Column(db.DateTime)
    needinfo = db.Column(db.Boolean(create_constraint=True, name='needinfo_bool'))
    needinfo_requestee = db.Column(db.String(1024), unique=False)
    last_whiteboard_change = db.Column(db.DateTime)
    """The time when we detected a blocker/FE keyword change. It doesn't really mark *any*
    whiteboard change, just for those tracked keywords."""
    proposed_blocker = db.Column(db.Boolean(create_constraint=True, name='proposed_blocker_bool'))
    proposed_fe = db.Column(db.Boolean(create_constraint=True, name='proposed_fe_bool'))
    rejected_blocker = db.Column(db.Boolean(create_constraint=True, name='rejected_blocker_bool'))
    rejected_fe = db.Column(db.Boolean(create_constraint=True, name='rejected_fe_bool'))
    accepted_blocker = db.Column(db.Boolean(create_constraint=True, name='accepted_blocker_bool'))
    accepted_0day = db.Column(db.Boolean(create_constraint=True, name='accepted_0day_bool'))
    accepted_prevrel = db.Column(db.Boolean(create_constraint=True, name='accepted_prevrel_bool'))
    accepted_fe = db.Column(db.Boolean(create_constraint=True, name='accepted_fe_bool'))
    prioritized = db.Column(db.Boolean(create_constraint=True, name='prioritized_bool'))
    milestone_id = db.Column(db.Integer, db.ForeignKey('milestone.id'))
    milestone: Optional['model_milestone.Milestone'] = db.relationship(
        'Milestone', back_populates='bugs')
    discussion_link = db.Column(db.String, unique=False)
    votes: str = db.Column(db.Text, default='{}', nullable=False)
    """A JSON representation of dict of all BugVoteTrackers and corresponding output of
    `BugVoteTracker.enumerate_votes()` with comment ids removed:
    ```
    {"tracker_name": {"-1": ["nick1", "nick2"],
                      "0":  [],
                      "+1": ["nick3"]}
     }
    ```"""
    _depends_on: str = db.Column('depends_on', db.Text, default='[]', nullable=False)
    """A JSON list of bug numbers which this bug depends on"""
    updates: list['model_update.Update'] = db.relationship(
        'Update', secondary=models.update_fixes, back_populates='bugs', lazy='dynamic',
        order_by=(model_update.Update.status.desc(), model_update.Update.pending.desc()))

    def __init__(self,
                 bugid: Optional[int],
                 url: Optional[str],
                 summary: Optional[str],
                 status: Optional[str],
                 component: Optional[str],
                 milestone: Optional['model_milestone.Milestone'],
                 active: Optional[bool],
                 needinfo: Optional[bool],
                 needinfo_requestee: Optional[str],
                 last_whiteboard_change: Optional[datetime.datetime] = datetime.datetime.utcnow(),
                 last_bug_sync: Optional[datetime.datetime] = datetime.datetime.now(),
                 depends_on: Optional[list[int]] = None) -> None:
        self.bugid = bugid
        self.url = url
        self.summary = summary
        self.status = status
        self.component = component
        self.active = active
        self.needinfo = needinfo
        self.needinfo_requestee = needinfo_requestee
        self.milestone = milestone
        self.proposed_blocker = False
        self.accepted_blocker = False
        self.accepted_0day = False
        self.accepted_prevrel = False
        self.rejected_blocker = False
        self.proposed_fe = False
        self.accepted_fe = False
        self.rejected_fe = False
        self.prioritized = False
        self.last_whiteboard_change = last_whiteboard_change
        self.last_bug_sync = last_bug_sync
        self.votes = "{}"
        self.depends_on = depends_on or []

    @property
    def depends_on(self) -> list[int]:
        """A list of bug numbers (int) which this bug depends on"""
        return json.loads(self._depends_on)

    @depends_on.setter
    def depends_on(self, value: Optional[list[int]]) -> None:
        self._depends_on = json.dumps(value or [])

    def __repr__(self):
        return '<bug %d: %s>' % (self.bugid, self.summary)

    def update(self, buginfo: dict, tracker_type: str, milestone: 'model_milestone.Milestone'
               ) -> None:
        """Update this Bug with new values. See `Bug.from_data()` for the description of
        arguments.
        """
        self.url = buginfo['url']
        self.summary = buginfo['summary']
        self.status = buginfo['status']
        self.component = buginfo['component']
        self.last_bug_sync = buginfo['last_change_time']
        self.active = buginfo['active']
        self.needinfo = buginfo['needinfo']
        self.needinfo_requestee = buginfo['needinfo_requestee']
        self.depends_on = buginfo['depends_on']
        if tracker_type == 'Blocker':
            if self.proposed_blocker != buginfo['proposed'] or \
               self.accepted_blocker != buginfo['accepted'] or \
               self.rejected_blocker != buginfo['rejected']:
                self.last_whiteboard_change = datetime.datetime.utcnow()
            self.proposed_blocker = buginfo['proposed']
            self.accepted_0day = buginfo['0day']
            self.accepted_prevrel = buginfo['prevrel']
            self.accepted_blocker = buginfo['accepted']
            self.rejected_blocker = buginfo['rejected']
        elif tracker_type == 'FreezeException':
            if self.proposed_fe != buginfo['proposed'] or \
               self.accepted_fe != buginfo['accepted'] or \
               self.rejected_fe != buginfo['rejected']:
                self.last_whiteboard_change = datetime.datetime.utcnow()
            self.proposed_fe = buginfo['proposed']
            self.accepted_fe = buginfo['accepted']
            self.rejected_fe = buginfo['rejected']
        self.prioritized = tracker_type == 'PrioritizedBug'
        if self.milestone and milestone.id != self.milestone.id:
            raise ValueError('Bugs with the same bugid and different milestones must be created'
                             'as separate instances!')

    @classmethod
    def from_data(cls, buginfo: dict[str, Any], milestone: 'model_milestone.Milestone',
                  tracker_type: str) -> 'Bug':
        """Create a `Bug` instance from the provided arguments

        :param buginfo: for a definition of this dict, see `BugSync.extract_information()`
        :param milestone: a `Milestone` instance
        :param tracker_type: one of 'Blocker', 'FreezeException' or 'PrioritizedBug'
        """
        newbug = Bug(buginfo['bug_id'], '', '', '', '', milestone, buginfo['active'], False, '')
        newbug.update(buginfo, tracker_type, milestone)
        return newbug
