# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

"""ORM class for release milestone info stored in the database"""

from typing import Any, Optional

from blockerbugs import db, BaseModel, models
from blockerbugs.models import bug, criterion, update
from blockerbugs.models import release as model_release


class Milestone(BaseModel):
    """A `Milestone` represents e.g. a Beta or a Final milestone, in relation to a certain
    `Release` (like 35).
    """

    id = db.Column(db.Integer, primary_key=True)
    release_id = db.Column(db.Integer, db.ForeignKey('release.id'))
    release: 'Optional[model_release.Release]' = db.relationship(
        'Release', back_populates='milestones')
    version = db.Column(db.String(80))
    """E.g. 'beta' or 'final'"""
    name = db.Column(db.String(80), unique=True)
    """A "pretty" name for the milestone, usually created together with the release number,
    e.g. '35-beta'"""
    blocker_tracker = db.Column(db.Integer, unique=True)
    """A bugzilla ticket number representing the tracker for blockers"""
    fe_tracker = db.Column(db.Integer, unique=True)
    """A bugzilla ticket number representing the tracker for freeze exceptions"""
    active = db.Column(db.Boolean(create_constraint=True, name='active_bool'))
    """Only active milestones are synced and displayed"""
    last_bug_sync = db.Column(db.DateTime)
    current = db.Column(db.Boolean(create_constraint=True, name='current_bool'))
    """Current milestone is the most relevant one currently. Usually it is the nearest milestone
    in the future. There should be at most one milestone marked as current."""
    bugs: list['bug.Bug'] = db.relationship('Bug', back_populates='milestone', lazy='dynamic')
    updates: list['update.Update'] = db.relationship('Update', secondary=models.update_milestones,
                                                     back_populates='milestones')
    criteria: list['criterion.Criterion'] = db.relationship('Criterion', back_populates='milestone',
                                                            lazy='dynamic')
    succeeds_id = db.Column(db.Integer, db.ForeignKey('milestone.id'), nullable=True)
    succeeds: 'Milestone' = db.relationship('Milestone', back_populates='succeeded_by',
                                            lazy='dynamic')
    succeeded_by: 'Milestone' = db.relationship('Milestone', back_populates='succeeds',
                                                remote_side=[id], uselist=False)

    def __init__(self,
                 release: Optional['model_release.Release'],
                 version: Optional[str],
                 blocker_tracker: Optional[int],
                 fe_tracker: Optional[int],
                 name: Optional[str],
                 active: Optional[bool] = False,
                 current: Optional[bool] = False,
                 succeeds: Optional['Milestone'] = None) -> None:
        self.release = release
        self.version = version
        self.blocker_tracker = blocker_tracker
        self.fe_tracker = fe_tracker
        self.name = name
        self.active = active
        self.last_bug_sync = None
        self.current = current
        if succeeds:
            self.succeeds = succeeds

    def __repr__(self) -> str:
        number = self.release.number if self.release else -1
        return 'milestone: %d-%s' % (number, self.version)

    def simple(self) -> dict[str, Any]:
        """Return a dict with basic Milestone information"""

        return {'id': self.id,
                'release': self.release.number if self.release else -1,
                'version': self.version,
                'name': self.name,
                'blocker_tracker': self.blocker_tracker,
                'fe_tracker': self.fe_tracker,
                'current': self.current,
                }
