# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

"""Database model for releases"""

from datetime import datetime
from typing import Optional

from blockerbugs import db, BaseModel
from blockerbugs.models import milestone, update


class Release(BaseModel):
    """This represents an OS release, primarily determined by its number."""

    id = db.Column(db.Integer, primary_key=True)
    number = db.Column(db.Integer)
    """E.g. 35 (for Fedora 35)"""
    active = db.Column(db.Boolean(create_constraint=True, name='active_bool'))
    """Only active releases are synced and displayed"""
    discussions_closed = db.Column(db.Boolean(create_constraint=True,
                                              name='discussions_closed_bool'))
    """This is used for past (historical) releases. If this is `True`, all discussion tickets (as in
    `Bug.discussion_link`) have been closed ("cleaned up")."""
    last_update_sync: datetime = db.Column(db.DateTime)
    milestones: list['milestone.Milestone'] = db.relationship('Milestone', back_populates='release',
                                                              lazy='dynamic')
    updates: list['update.Update'] = db.relationship('Update', back_populates='release',
                                                     lazy='dynamic')

    def __init__(self, number: int, active: Optional[bool] = True,
                 discussions_closed: Optional[bool] = False) -> None:
        self.number = number
        self.active = active
        self.discussions_closed = discussions_closed
        # initialize last_update_time so there's at least something there
        self.last_update_sync = datetime(1990, 1, 1, 1, 0, 0, 0)

    def __str__(self) -> str:
        return 'release: %d' % (self.number)
