# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

"""Database model for Bodhi updates"""

from blockerbugs import db, BaseModel, models
from blockerbugs.models import bug, milestone
from blockerbugs.models import release as model_release


class Update(BaseModel):
    id = db.Column(db.Integer, primary_key=True)
    title = db.Column(db.Text, unique=False)
    url = db.Column(db.Text, unique=False)
    karma = db.Column(db.Integer, unique=False)
    stable_karma = db.Column(db.Integer, unique=False)
    status = db.Column(db.String(80), unique=False)
    request = db.Column(db.String(80), unique=False, nullable=True)
    pending = db.Column(db.Boolean(create_constraint=True, name='pending_bool'), unique=False)
    date_submitted = db.Column(db.DateTime)
    date_pushed_testing = db.Column(db.DateTime, nullable=True)
    date_pushed_stable = db.Column(db.DateTime, nullable=True)
    date_obsoleted = db.Column(db.DateTime, nullable=True)
    bugs: list['bug.Bug'] = db.relationship(
        'Bug', secondary=models.update_fixes, back_populates='updates')
    release_id = db.Column(db.Integer, db.ForeignKey('release.id'))
    release: 'model_release.Release' = db.relationship('Release', back_populates='updates')
    milestones: list['milestone.Milestone'] = db.relationship(
        'Milestone', secondary=models.update_milestones, back_populates='updates')

    def __init__(self, title, url, karma, status, bugs, release,
                 milestones, stable_karma=3, date_submitted=None):
        self.title = title
        self.url = url
        self.karma = karma
        self.status = status
        self.request = None
        self.bugs = bugs
        self.date_submitted = date_submitted
        self.release = release
        if milestones:
            self.milestones = milestones
        self.date_pushed_testing = None
        self.date_pushed_stable = None
        self.date_obsoleted = None

    def __str__(self):
        return 'update: %s' % (self.title)

    def sync(self, updateinfo):
        self.title = updateinfo['title']
        self.url = updateinfo['url']
        self.karma = updateinfo['karma']
        self.stable_karma = updateinfo['stable_karma']
        self.status = updateinfo['status']
        self.request = updateinfo['request']
        self.date_submitted = updateinfo['date_submitted']
        if updateinfo['date_pushed_testing']:
            self.date_pushed_testing = updateinfo['date_pushed_testing']
        if updateinfo['date_pushed_stable']:
            self.date_pushed_stable = updateinfo['date_pushed_stable']
        self.pending = updateinfo['pending']
        current_bugkeys = [(currentbug.bugid, currentbug.milestone) for currentbug in self.bugs]
        for bugid in updateinfo['bugs']:
            newbugs = bug.Bug.query.filter_by(bugid=bugid).all()
            for newbug in newbugs:
                if not (newbug.bugid, newbug.milestone) in current_bugkeys:
                    self.bugs.append(newbug)
                    current_bugkeys.append((newbug.bugid, newbug.milestone))
                if newbug.milestone and newbug.milestone not in self.milestones:
                    self.milestones.append(newbug.milestone)

    @classmethod
    def from_data(cls, updateinfo, release):
        newupdate = Update(updateinfo['title'], '', 0, '', [], release, None)
        newupdate.sync(updateinfo)
        return newupdate
