# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

"""Sync bug information from Bugzilla"""

from blockerbugs.models.bug import Bug
from blockerbugs.models.milestone import Milestone
from blockerbugs.util import bz_interface
from datetime import datetime
from sqlalchemy import or_
import flask_sqlalchemy
from bugzilla.bug import Bug as bzBug
import logging
from typing import Any, Optional


class BugSync():
    """Main class for syncing existing and retrieving new blocker (and other) bugs"""

    def __init__(self, db: flask_sqlalchemy.SQLAlchemy,
                 bzinterface: Optional[bz_interface.BlockerBugs] = None, url: Optional[str] = None
                 ) -> None:
        """:param url: Bugzilla API url"""
        self.db: flask_sqlalchemy.SQLAlchemy = db
        self.log: logging.Logger = logging.getLogger('bug_sync')
        self.bzinterface: bz_interface.BlockerBugs = (bzinterface or
                                                      bz_interface.BlockerBugs(url=url))

    def extract_information(self, bug: bzBug, tracker_type: str) -> dict[str, Any]:
        """Create a dict with extracted Bug information. See the code to learn the dict keyvals.

        :param tracker_type: one of 'Blocker', 'FreezeException' or 'PrioritizedBug'
        """
        buginfo: dict[str, Any] = {}
        bug_whiteboard = bug.whiteboard

        # get information about accepted/rejected
        buginfo['0day'] = 'Accepted0Day' in bug_whiteboard
        buginfo['prevrel'] = 'AcceptedPreviousRelease' in bug_whiteboard
        buginfo['accepted'] = ('Accepted%s' % tracker_type) in bug_whiteboard
        buginfo['rejected'] = ('Rejected%s' % tracker_type) in bug_whiteboard
        buginfo['proposed'] = not (buginfo['rejected'] or buginfo['accepted'] or
                                   buginfo['prevrel'] or buginfo['0day'])

        # determine whether or not the bug is active (not CLOSED)
        buginfo['active'] = bug.is_open

        # basic info
        buginfo['bug_id'] = bug.bug_id
        buginfo['component'] = bug.component
        buginfo['summary'] = bug.summary
        buginfo['status'] = bug.status
        buginfo['url'] = bug.weburl
        buginfo['needinfo'] = False
        buginfo['needinfo_requestee'] = ''
        buginfo['depends_on'] = bug.dependson

        # needinfo flag
        if hasattr(bug, 'flags'):
            needinfos = []
            for flag in bug.flags:
                if flag['name'] == 'needinfo' and flag['status'] == '?' and flag['is_active']:
                    buginfo['needinfo'] = True
                    try:
                        needinfos.append(flag['requestee'])
                    except KeyError:
                        pass
            buginfo['needinfo_requestee'] = ', '.join(needinfos)

        # according to bugzilla's docs, this is always in UTC
        last_change_time = bug.last_change_time.timetuple()
        buginfo['last_change_time'] = datetime(last_change_time.tm_year,
                                               last_change_time.tm_mon,
                                               last_change_time.tm_mday,
                                               last_change_time.tm_hour,
                                               last_change_time.tm_min,
                                               last_change_time.tm_sec)
        return buginfo

    def get_tracker_id(self, milestone: Milestone, tracker_type: str) -> Optional[int]:
        """Return the bug number which serves as a tracker for the specific tracker type in the
        specified Milestone.

        :param tracker_type: one of 'Blocker', 'FreezeException' or 'PrioritizedBug'
        :returns: the bug number, or `None` (for `tracker_type == 'PrioritizedBug'`)
        """
        if tracker_type == 'Blocker':
            return milestone.blocker_tracker
        elif tracker_type == 'FreezeException':
            return milestone.fe_tracker
        elif tracker_type == 'PrioritizedBug':
            return None
        else:
            raise ValueError(f'Invalid tracker_type: {tracker_type}')

    def update_milestone(self, milestone: Milestone, tracker_type: str,
                         lastupdate: Optional[datetime] = None) -> None:
        """Retrieve all new bugs for a `tracker_type` in a `milestone` since `lastupdate` (or ever,
        if not specified) from Bugzilla. Find bugs which already exist in the database and update
        them, and create new database entries for those yet untracked

        :param tracker_type: one of 'Blocker', 'FreezeException' or 'PrioritizedBug'
        """
        if tracker_type in ['Blocker', 'FreezeException']:
            trackerid = self.get_tracker_id(milestone, tracker_type)
            assert trackerid
            self.log.info('Updating %s for Fedora %d %s (%d)' % (
                tracker_type, milestone.release.number if milestone.release else -1,
                milestone.version, trackerid))
            bugs = self.bzinterface.query_tracker(trackerid, last_update=lastupdate)
        elif tracker_type == 'PrioritizedBug':
            self.log.info('Updating %s for Fedora %d %s' % (
                tracker_type, milestone.release.number if milestone.release else -1,
                milestone.version))
            bugs = self.bzinterface.query_prioritized()
        else:
            raise ValueError(f'Invalid tracker_type: {tracker_type}')
        self.log.debug('Found bugs: %s' % str(bugs))

        # sanity check - all retrieved bugs should be unique
        bugids = [bug.bug_id for bug in bugs]
        assert len(bugids) == len(set(bugids))

        # update existing bugs, add new ones
        for bug in bugs:
            oldbug = Bug.query.filter_by(bugid=bug.bug_id, milestone=milestone).first()
            buginfo = self.extract_information(bug, tracker_type)
            if oldbug:
                self.log.debug('Updating bug %d' % buginfo['bug_id'])
                oldbug.update(buginfo, tracker_type, milestone)
                self.db.session.add(oldbug)
            else:
                self.log.debug('Adding new bug %d' % buginfo['bug_id'])
                newbug = Bug.from_data(buginfo, milestone, tracker_type)
                self.db.session.add(newbug)

        self.db.session.commit()

    def cleanup_milestone(self, milestone: Milestone, tracker_type: str) -> None:
        """'Clean up' all bugs under `tracker_type` in `milestone`, which are not currently blocking
        the relevant tracker ticket. Cleaning up means setting all those bugs as rejected
        (blocker/FE/prioritized) and unsetting proposed/accepted fields. That will make those bugs
        disappear from the UI and ignored in future BZ queries (unless the tracker starts depending
        on them again). So for example this might process and 'clean up' all F34-beta blockers,
        which no longer block the F34BetaBlocker tracker ticket.

        :param tracker_type: one of 'Blocker', 'FreezeException' or 'PrioritizedBug'
        """
        # retrieve all *current* bugs blocking the tracker
        if tracker_type in ['Blocker', 'FreezeException']:
            trackerid = self.get_tracker_id(milestone, tracker_type)
            assert trackerid
            self.log.info('Cleaning up %s for Fedora %d %s (%d)' % (
                tracker_type, milestone.release.number if milestone.release else -1,
                milestone.version, trackerid))
            tracker_depends = set(self.bzinterface.get_deps(trackerid))
            self.log.debug("Current dependencies for %d: %s" % (trackerid, str(tracker_depends)))
        elif tracker_type == 'PrioritizedBug':
            self.log.info('Cleaning up %s for Fedora %d %s' % (
                tracker_type, milestone.release.number if milestone.release else -1,
                milestone.version))
            tracker_depends = set()
        else:
            raise ValueError(f'Invalid tracker_type: {tracker_type}')

        # get all matching bugs from the DB
        blockers: list[Bug]
        if tracker_type == 'FreezeException':
            blockers = Bug.query.filter_by(milestone=milestone).filter(
                or_(Bug.proposed_fe == True,  # noqa: E712
                    Bug.accepted_fe == True)).all()
        elif tracker_type == 'Blocker':
            blockers = Bug.query.filter_by(milestone=milestone).filter(
                or_(Bug.proposed_blocker == True,  # noqa: E712
                    Bug.accepted_blocker == True,
                    Bug.accepted_0day == True,
                    Bug.accepted_prevrel == True)).all()
        elif tracker_type == 'PrioritizedBug':
            blockers = Bug.query.filter_by(milestone=milestone).filter(
                Bug.prioritized == True).all()  # noqa: E712

        # clean up bugs which no longer block the tracker
        for blocker in blockers:
            if blocker.bugid not in tracker_depends:
                if tracker_type == 'FreezeException':
                    blocker.proposed_fe = False
                    blocker.accepted_fe = False
                    blocker.rejected_fe = True
                elif tracker_type == 'Blocker':
                    blocker.proposed_blocker = False
                    blocker.accepted_blocker = False
                    blocker.accepted_0day = False
                    blocker.accepted_prevrel = False
                    blocker.rejected_blocker = True
                elif tracker_type == 'PrioritizedBug':
                    blocker.prioritized = False
                self.db.session.add(blocker)
        self.db.session.commit()

    def update_active(self, full_sync: Optional[bool] = False) -> None:
        """Update bugs from all active milestones.

        :param full_sync: Perform a full sync (retrieve all data) or a quick sync (retrieve only
                          new data since the last update).
        """
        starttime = datetime.utcnow()
        active_milestones = Milestone.query.filter_by(active=True).all()
        self.log.info("Starting blocker sync")

        # for each active milestone, update the tracker to remove any bugs that
        # no longer block the tracker and pull in any new bugs
        for milestone in active_milestones:
            last_sync_time = milestone.last_bug_sync
            if full_sync:
                last_sync_time = None

            self.log.info("Syncing blockers for %s" % milestone)
            self.cleanup_milestone(milestone, 'Blocker')
            self.update_milestone(milestone, 'Blocker', last_sync_time)

            self.log.info("Syncing FE for %s" % milestone)
            self.cleanup_milestone(milestone, 'FreezeException')
            self.update_milestone(milestone, 'FreezeException', last_sync_time)

            self.log.info("Syncing PrioritizedBug for %s" % milestone)
            self.cleanup_milestone(milestone, 'PrioritizedBug')
            self.update_milestone(milestone, 'PrioritizedBug', None)

            # update the last query time for the milestone
            milestone.last_bug_sync = starttime
            self.db.session.add(milestone)
            self.db.session.commit()
