from datetime import datetime
import json
import copy
import http.client as httplib

import mock

from blockerbugs import db
from blockerbugs import app
from blockerbugs.models.bug import Bug
from blockerbugs.models.build import Build
from blockerbugs.models.userinfo import UserInfo
from testing.test_controllers import add_release, add_milestone, \
    add_bug, add_update
from blockerbugs.controllers.api import api, errors
from blockerbugs.controllers.api.api import _get_bugtypes, _UNKNOWN_BUG_SVG_TEXT, _BUG_CLOSED
from blockerbugs.util import pagure_bot


class TestRestAPI(object):
    @classmethod
    def setup_class(cls):
        db.session.rollback()
        db.drop_all()
        db.create_all()
        cls.client = app.test_client()
        cls.release = add_release(99)
        cls.milestone = add_milestone(cls.release, 'final', 100, 101,
                                      '99-final', True)
        cls.milestone.current = True

        cls.milestone2 = add_milestone(cls.release, 'beta', 200, 201,
                                       '99-beta')
        bug1 = add_bug(9000, 'testbug1', cls.milestone)
        bug1.accepted_fe = True
        bug1.status = 'CLOSED'
        bug2 = add_bug(9002, 'testbug2', cls.milestone)
        bug2.accepted_blocker = False
        bug2.proposed_fe = True
        bug3 = add_bug(9003, 'testbug3', cls.milestone)
        bug3.accepted_blocker = False
        bug3.proposed_fe = True
        bug4 = add_bug(9003, 'testbug3', cls.milestone2) # same bugid and summary as bug3 is intentional
        bug4.accepted_blocker = True
        bug4.proposed_fe = False
        bug4.status = 'CLOSED'
        cls.update_pending_stable = add_update(u'test-pending-stable.fc99',
                                               u'stable', [bug1], cls.release,
                                               [cls.milestone, cls.milestone2])
        cls.update_pending_stable.date_submitted = datetime(1990, 1, 1)
        cls.update_pending_stable.pending = True
        cls.update_testing2 = add_update(u'test-testing2.fc99', u'testing',
                                         [bug2],
                                         cls.release, [cls.milestone])
        build = Build()
        build.koji_id = 33
        build.nvr = 'libofx-0.9.9-1.fc20'

        cls.webhook_data = {
            'msg': {
                'issue': {
                    'id': 6666,
                    'status': 'Open',
                },
            },
            'topic': 'issue.comment.added',
        }

        cls.api_user = UserInfo('api_user')
        passwd = cls.api_user.generate_api_key()
        cls.headers = {'X-Auth-User': 'api_user', 'X-Auth-Key': passwd}
        db.session.add(cls.api_user)
        db.session.commit()

    @classmethod
    def teardown_class(cls):
        db.session.rollback()
        db.drop_all()

    # === /api/v0/milestones ===

    def test_unknown_release(self):
        url = '/api/v0/milestones/9999/unknown/bugs'
        resp = self.client.get(url)
        assert resp.status_code == httplib.NOT_FOUND
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.NoSuchObjectError.code
        assert 'Release' in error['message']

    def test_unknown_milestone(self):
        url = '/api/v0/milestones/99/unknown/bugs'
        resp = self.client.get(url)
        assert resp.status_code == httplib.NOT_FOUND
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.NoSuchObjectError.code
        assert 'Milestone' in error['message']

    def test_list_all_bugs(self):
        url = '/api/v0/milestones/99/final/bugs'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 3
        bug = data[0]
        assert bug['bugid'] == 9000
        assert bug['url'] == 'https://bugzilla.redhat.com/show_bug.cgi?id=9000'
        assert bug['summary'] == 'testbug1'
        assert bug['component'] == 'testcomponent'
        assert bug['active']
        assert set(bug['type']) == set(('accepted_blocker', 'accepted_fe'))

    def test_list_accepted_blocker_bugs(self):
        url = '/api/v0/milestones/99/final/bugs?bugtype=accepted_blocker&'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 1
        assert data[0]['bugid'] == 9000

    def test_list_all_updates(self):
        url = '/api/v0/milestones/99/final/updates'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 2
        update = data[-1]
        assert update['title'] == u'test-pending-stable.fc99'
        assert update['url'] == u'http://localhost/update'
        assert update['karma'] == 1
        assert update['stable_karma'] == 3
        assert update['status'] == u'stable'
        assert update['pending']
        assert update['bugs'][0]['bugid'] == 9000
        assert set(update['bugs'][0]['type']) == set(('accepted_blocker',
                                                  'accepted_fe'))
        assert update['release'] == 99
        assert len(update['milestones']) == 2
        assert {'version': 'final', 'release': 99} in update['milestones']
        assert {'version': 'beta', 'release': 99} in update['milestones']

    def test_list_updates_with_proposed_fe_bugs(self):
        url = '/api/v0/milestones/99/final/updates?bugtype=proposed_fe&'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert len(data) == 1
        update = data[0]
        assert update['title'] == u'test-testing2.fc99'

    def test_bad_bugtype_list_bugs(self):
        url = '/api/v0/milestones/99/final/updates?bugtype=foo&'
        resp = self.client.get(url)
        assert resp.status_code == httplib.BAD_REQUEST
        error = json.loads(resp.data)['error']
        assert error['code'] == errors.InvalidArgumentError.code
        assert 'bugtype' in error['message']

    def test_get_current_milestone(self):
        url = '/api/v0/milestones/current'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = json.loads(resp.data)
        assert data['name'] == self.milestone.name

    # === /api/v0/bugimg ===

    def test_get_bugimg(self):
        url = '/api/v0/bugimg/9002'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = str(resp.data)

        bug2 = db.session.query(Bug).filter_by(bugid = 9002).first()

        assert self.milestone.name in data
        for bugtype in _get_bugtypes(bug2):
            assert bugtype in data

    def test_get_bugimg_wrong_bugid(self):
        url = '/api/v0/bugimg/90210'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = str(resp.data)

        assert self.milestone.name not in data
        assert _UNKNOWN_BUG_SVG_TEXT in data

    def test_get_bugimg_open_bug(self):
        url = '/api/v0/bugimg/9002'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = str(resp.data)

        assert _BUG_CLOSED not in data

    def test_get_bugimg_closed_bug(self):
        url = '/api/v0/bugimg/9000'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = str(resp.data)

        assert _BUG_CLOSED in data

    def test_get_bugimg_multiple_bugs_unsynced_status(self):
        url = '/api/v0/bugimg/9003'
        resp = self.client.get(url)
        assert resp.status_code == httplib.OK
        data = str(resp.data)

        assert '99-beta' in data
        assert '99-final' in data
        assert _BUG_CLOSED not in data

    # === /api/v0/webhook ===

    def test_webhook_bot_disabled(self, monkeypatch):
        url = '/api/v0/webhook'
        mock_webhook_handler = mock.MagicMock()
        monkeypatch.setattr(pagure_bot, 'webhook_handler', mock_webhook_handler)
        monkeypatch.setattr(api, 'check_signature', mock.MagicMock(return_value=True))
        monkeypatch.setitem(app.config, 'PAGURE_BOT_ENABLED', False)

        resp = self.client.post(url, json=self.webhook_data)
        assert resp.status_code == httplib.OK
        respdata = json.loads(resp.data)

        assert not mock_webhook_handler.called
        assert respdata['msg'] == 'Pagure bot disabled, ignoring request'

    def test_webhook_bad_signature(self, monkeypatch):
        # If we don't mock the signature, it doesn't match
        url = '/api/v0/webhook'
        mock_webhook_handler = mock.MagicMock()
        monkeypatch.setattr(pagure_bot, 'webhook_handler', mock_webhook_handler)

        resp = self.client.post(url, json=self.webhook_data)
        assert resp.status_code == httplib.OK
        respdata = json.loads(resp.data)
        print(respdata['msg'])

        assert not mock_webhook_handler.called
        assert respdata['msg'] == 'Wrong signature, ignoring.'

    def test_webhook_wrong_topic(self, monkeypatch):
        url = '/api/v0/webhook'
        mock_webhook_handler = mock.MagicMock()
        monkeypatch.setattr(pagure_bot, 'webhook_handler', mock_webhook_handler)
        monkeypatch.setattr(api, 'check_signature', mock.MagicMock(return_value=True))
        monkeypatch.setitem(self.webhook_data, 'topic', 'test.invalid.topic')

        resp = self.client.post(url, json=self.webhook_data)
        assert resp.status_code == httplib.OK
        respdata = json.loads(resp.data)

        assert not mock_webhook_handler.called
        assert respdata['msg'].startswith('Ignoring message with topic')

    def test_webhook_missing_fields(self, monkeypatch):
        url = '/api/v0/webhook'
        mock_webhook_handler = mock.MagicMock()
        monkeypatch.setattr(pagure_bot, 'webhook_handler', mock_webhook_handler)
        monkeypatch.setattr(api, 'check_signature', mock.MagicMock(return_value=True))

        for field in ['id', 'status']:
            webhook_data = copy.deepcopy(self.webhook_data)
            del webhook_data['msg']['issue'][field]

            resp = self.client.post(url, json=webhook_data)
            assert resp.status_code == httplib.OK
            respdata = json.loads(resp.data)

            assert not mock_webhook_handler.called
            assert respdata['msg'].startswith('Unable to parse received message')

        for field in ['id', 'status']:
            for value in ['', None]:
                webhook_data = copy.deepcopy(self.webhook_data)
                webhook_data['msg']['issue'][field] = value

                resp = self.client.post(url, json=webhook_data)
                assert resp.status_code == httplib.OK
                respdata = json.loads(resp.data)

                assert not mock_webhook_handler.called
                assert respdata['msg'].startswith('Unable to parse received message')

    def test_webhook_closed_issue(self, monkeypatch):
        url = '/api/v0/webhook'
        mock_webhook_handler = mock.MagicMock()
        monkeypatch.setattr(pagure_bot, 'webhook_handler', mock_webhook_handler)
        monkeypatch.setattr(api, 'check_signature', mock.MagicMock(return_value=True))
        monkeypatch.setitem(self.webhook_data['msg']['issue'], 'status', 'Closed')

        resp = self.client.post(url, json=self.webhook_data)
        assert resp.status_code == httplib.OK
        respdata = json.loads(resp.data)

        assert not mock_webhook_handler.called
        assert respdata['msg'] == 'Ignoring a closed issue'

    def test_webhook_executed(self, monkeypatch):
        url = '/api/v0/webhook'
        mock_webhook_handler = mock.MagicMock()
        monkeypatch.setattr(pagure_bot, 'webhook_handler', mock_webhook_handler)
        monkeypatch.setattr(api, 'check_signature', mock.MagicMock(return_value=True))

        resp = self.client.post(url, json=self.webhook_data)
        assert resp.status_code == httplib.OK
        respdata = json.loads(resp.data)

        assert mock_webhook_handler.call_count == 1
        assert mock_webhook_handler.call_args.args == (self.webhook_data['msg']['issue']['id'],)
        assert respdata['msg'] == 'Message successfully parsed'
