from munch import Munch
from copy import deepcopy
from blockerbugs.util.bug_sync import BugSync
from mock import MagicMock
import datetime

basicbug = Munch({'bug_id': 123456,
                  'weburl': 'https://bugzilla.redhat.com/show_bug.cgi?id=123456',
                  'component': 'randomcomponent',
                  'status': 'NEW',
                  'is_open': True,
                  'summary': 'this is a test bug',
                  'whiteboard': '',
                  'dependson': [333],
                  'last_change_time': datetime.datetime.utcnow(),
                  'flags': [{'name': 'needinfo', 'status': '?', 'is_active': True, 'requestee': 'John Doe'}]
})


class TestSyncExtractInformation(object):
    def setup_method(self, method):
        self.testbug = deepcopy(basicbug)
        self.testsync = BugSync(None, MagicMock())

    def test_nth_isaccepted(self):
        self.testbug.whiteboard = 'AcceptedFreezeException'
        ref_isaccepted = True

        buginfo = self.testsync.extract_information(self.testbug,
                                                    'FreezeException')

        assert buginfo['accepted'] == ref_isaccepted

    def test_nth_isrejected(self):
        self.testbug.whiteboard = 'RejectedFreezeException'
        ref_isrejected = True

        buginfo = self.testsync.extract_information(self.testbug,
                                                    'FreezeException')

        assert buginfo['rejected'] == ref_isrejected

    def test_nth_isproposed(self):
        self.testbug.whiteboard = ''
        ref_isproposed = True

        buginfo = self.testsync.extract_information(self.testbug,
                                                    'FreezeException')

        assert buginfo['proposed'] == ref_isproposed

    def test_nth_isproposed_notrejected(self):
        self.testbug.whiteboard = ''
        ref_isrejected = False

        buginfo = self.testsync.extract_information(self.testbug,
                                                    'FreezeException')

        assert buginfo['rejected'] == ref_isrejected

    def test_nth_isproposed_notaccepted(self):
        self.testbug.whiteboard = ''
        ref_isaccepted = False

        buginfo = self.testsync.extract_information(self.testbug,
                                                    'FreezeException')

        assert buginfo['accepted'] == ref_isaccepted

    def test_blocker_isaccepted(self):
        self.testbug.whiteboard = 'AcceptedBlocker'
        ref_isaccepted = True

        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['accepted'] == ref_isaccepted

    def test_blocker_isaccepted_0day(self):
        self.testbug.whiteboard = 'Accepted0Day'
        ref_isaccepted = True

        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['0day'] == ref_isaccepted

    def test_blocker_isaccepted_prevrel(self):
        self.testbug.whiteboard = 'AcceptedPreviousRelease'
        ref_isaccepted = True

        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['prevrel'] == ref_isaccepted

    def test_blocker_isaccepted_notproposed(self):
        self.testbug.whiteboard = 'AcceptedBlocker'

        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['proposed'] == False

    def test_bugid(self):
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['bug_id'] == self.testbug['bug_id']

    def test_component(self):
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['component'] == self.testbug['component']

    def test_title(self):
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['summary'] == self.testbug['summary']

    def test_status(self):
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['status'] == self.testbug['status']

    def test_url(self):
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['url'] == self.testbug['weburl']

    def test_closed_inactive(self):
        self.testbug['status'] = 'CLOSED'
        self.testbug['is_open'] = False
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['status'] == self.testbug['status']
        assert buginfo['active'] is False

    def test_notclosed_active(self):
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['active'] is True

    def test_needinfo(self):
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['needinfo']
        assert buginfo['needinfo_requestee'] == 'John Doe'

    def test_multiple_needinfo(self):
        self.testbug.flags.append({'name': 'needinfo', 'status': '?', 'is_active': True, 'requestee': 'Jane Doe'})
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['needinfo']
        assert buginfo['needinfo_requestee'] == 'John Doe, Jane Doe'

    def test_needinfo_no_requestee(self):
        self.testbug.flags[0].pop('requestee')
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['needinfo']
        assert buginfo['needinfo_requestee'] == ''

    def test_multiple_needinfo_no_requestee(self):
        self.testbug.flags.append({'name': 'needinfo', 'status': '?', 'is_active': True})
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['needinfo']
        assert buginfo['needinfo_requestee'] == 'John Doe'

    def test_needinfo_empty(self):
        self.testbug.flags = []
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')

        assert buginfo['needinfo'] == False
        assert buginfo['needinfo_requestee'] == ''

    def test_dependson(self):
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')
        assert buginfo['depends_on'] == self.testbug['dependson']

    def test_dependson_empty(self):
        self.testbug.depends_on = []
        buginfo = self.testsync.extract_information(self.testbug, 'Blocker')
        assert buginfo['depends_on'] == self.testbug['dependson']
