import datetime

from blockerbugs.models.milestone import Milestone
from blockerbugs.models.release import Release
from blockerbugs.models.bug import Bug
from blockerbugs.models.update import Update
from blockerbugs import db
from blockerbugs import app
from blockerbugs.controllers import main


def add_release(number=99):
    test_release = Release(number)
    db.session.add(test_release)
    db.session.commit()
    return test_release


def add_milestone(release, version, blocker_tracker, fe_tracker, name,
                  active=False):
    test_milestone = Milestone(release, version, blocker_tracker, fe_tracker,
                               name, active)
    db.session.add(test_milestone)
    db.session.commit()
    return test_milestone


def add_bug(bugid, summary, milestone):
    test_bug = Bug(bugid,
                   'https://bugzilla.redhat.com/show_bug.cgi?id=%d' % bugid,
                   summary, 'NEW', 'testcomponent', milestone, True, True, 'John Doe')
    test_bug.accepted_blocker = True
    db.session.add(test_bug)
    db.session.commit()
    return test_bug


def add_update(title, status, bugs, release, milestones,
               url='http://localhost/update'):
    test_update = Update(title, url, 1, status, bugs,
                         release, milestones,
                         date_submitted=datetime.datetime.utcnow())
    test_update.stable_karma = 3
    db.session.add(test_update)
    db.session.commit()
    return test_update


def item_just_in(item, correct_containers, all_containers):
    """Check whether `item` is present in all of `correct_containers` (e.g. strings, lists, etc) and
    in none of other containers which are the result of `all_containers - correct_containers`.
    """
    for container in correct_containers:
        if item not in container:
            return False
    for container in all_containers:
        if container in correct_containers:
            continue
        if item in container:
            return False
    return True


def split_requests(rendered_requests: str) -> tuple[str, str, str, str, str]:
    """Take a rendered 'requests.txt' template, split it into parts and return them.

    :returns: (compose_blockers, compose_fes, push_blockers, push_fes, deps)
    """
    index_fp = rendered_requests.index('### FREEZE PUSH ###')
    try:
        index_deps = rendered_requests.index('### THE ABOVE MIGHT NOT BE COMPLETE')
    except ValueError:
        # deps section might not be present
        index_deps = len(rendered_requests)
    compose = rendered_requests[:index_fp]
    assert '### CANDIDATE COMPOSE ###' in compose
    push = rendered_requests[index_fp:index_deps]
    assert '### FREEZE PUSH ###' in push
    deps = rendered_requests[index_deps:]
    assert '### THE ABOVE MIGHT NOT BE COMPLETE' in deps or not deps
    index_fe = compose.index('== Freeze exceptions ==')
    compose_blockers = compose[:index_fe]
    assert '== Blockers ==' in compose_blockers
    compose_fes = compose[index_fe:]
    assert '== Freeze exceptions ==' in compose_fes
    index_fe = push.index('== Freeze exceptions ==')
    push_blockers = push[:index_fe]
    assert '== Blockers ==' in push_blockers
    push_fes = push[index_fe:]
    assert '== Freeze exceptions ==' in push_fes

    all_sections = (compose_blockers, compose_fes, push_blockers, push_fes, deps)
    assert len(rendered_requests) == len(''.join(all_sections))
    return all_sections


class TestEmptyDB(object):
    @classmethod
    def setup_class(cls):
        cls.client = app.test_client()
        db.session.rollback()
        db.drop_all()
        db.create_all()

    @classmethod
    def teardown_class(cls):
        db.session.rollback()
        db.drop_all()

    def test_index_returns_200code(self):
        rv = self.client.get('/')
        assert rv.status_code == 200


class TestNotSyncedMilestone(object):
    @classmethod
    def setup_class(cls):
        cls.client = app.test_client()
        db.session.rollback()
        db.drop_all()
        db.create_all()
        release = add_release(99)
        add_milestone(release, 'final', 100, 101, '99-final', True)
        add_milestone(release, 'beta', 200, 201, '99-beta')

    @classmethod
    def teardown_class(cls):
        db.session.rollback()
        db.drop_all()

    def test_not_synced_index(self):
        rv = self.client.get('/')
        assert rv.status_code == 200
        assert b'F99 final' in rv.data
        assert b'F99 beta' not in rv.data

    def test_display_buglist_returns_200code(self):
        rv = self.client.get('milestone/99/final/buglist')
        assert rv.status_code == 200

    def test_display_updates_returns_200code(self):
        rv = self.client.get('milestone/99/final/updates')
        assert rv.status_code == 200

    def test_display_blockers_returns_200code(self):
        rv = self.client.get('milestone/99/final/irc')
        assert rv.status_code == 200

    def test_display_info_returns_200code(self):
        rv = self.client.get('milestone/99/final/info')
        assert rv.status_code == 200


class TestSyncedMilestone(object):
    @classmethod
    def setup_class(cls):
        cls.client = app.test_client()
        db.session.rollback()
        db.drop_all()
        db.create_all()
        release = add_release(99)
        final_milestone = add_milestone(release, 'final', 100, 101,
                                        '99-final', True)
        add_milestone(release, 'beta', 200, 201, '99-beta')
        add_bug(9000, 'testbug1', final_milestone)
        add_bug(9001, 'testbug2', final_milestone)
        final_milestone.last_bug_sync = datetime.datetime.utcnow()
        db.session.commit()

    @classmethod
    def teardown_class(cls):
        db.session.rollback()
        db.drop_all()

    def test_display_index(self):
        rv = self.client.get('/')
        assert rv.status_code == 200
        assert b'F99 final' in rv.data
        assert b'F99 beta' not in rv.data

    def test_display_milestone_redirect(self):
        rv = self.client.get('milestone/99/final')
        assert rv.status_code == 302

    def test_display_buglist(self):
        rv = self.client.get('milestone/99/final/buglist')
        assert rv.status_code == 200
        assert b'9000' in rv.data
        assert b'testbug2' in rv.data

    def test_display_updates_returns_200code(self):
        rv = self.client.get('milestone/99/final/updates')
        assert rv.status_code == 200

    def test_display_irc_blockers(self):
        rv = self.client.get('milestone/99/final/irc')
        assert rv.status_code == 200
        assert b'9000' in rv.data
        assert b'testbug2' in rv.data

    def test_display_info(self):
        rv = self.client.get('milestone/99/final/info')
        assert rv.status_code == 200
        assert b'100' in rv.data
        assert b'101' in rv.data

    def test_display_stats_returns_200code(self):
        rv = self.client.get('milestone/99/final/stats')
        assert rv.status_code == 200


class TestGetFunctions(object):
    @classmethod
    def setup_class(cls):
        db.session.rollback()
        db.drop_all()
        db.create_all()
        cls.release = add_release(99)
        cls.milestone = add_milestone(cls.release, 'beta', 100, 101,
                                        '99-beta', True)
        bug1 = add_bug(9000, 'testbug1', cls.milestone)
        bug1.depends_on = [9090]
        bug2 = add_bug(9001, 'testbug2', cls.milestone)
        bug2.accepted_blocker = False
        # a bug that fixes a Final blocker
        cls.finalmile = add_milestone(cls.release, 'final', 200, 201, '99-final')
        finalbug = add_bug(9002, 'finalbug', cls.finalmile)
        # we add finalbug to this update for requests.txt testing:
        # it should *not* show in the generated request, the update
        # will be in the request but it should only list bug1, not
        # finalbug, as bug1 is for Beta and finalbug for Final
        cls.update_pending_stable = add_update(u'test-pending-stable.fc99', u'stable',
                            [bug1, finalbug], cls.release, [cls.milestone])
        cls.update_pending_stable.pending = True
        cls.update_pending_stable.request = 'stable'
        cls.update_stable1 = add_update(u'test-stable1.fc99', u'stable', [bug1],
                                        cls.release, [cls.milestone])
        cls.update_stable2 = add_update(u'test-stable2.fc99', u'stable', [bug2],
                                        cls.release, [cls.milestone])
        cls.update_pending_testing = add_update(u'test-pending-testing.fc99', u'testing', [bug1],
                            cls.release, [cls.milestone])
        cls.update_pending_testing.pending = True
        cls.update_testing1 = add_update(u'test-testing1.fc99', u'testing', [bug1],
                                        cls.release, [cls.milestone])
        cls.update_testing2 = add_update(u'test-testing2.fc99', u'testing', [bug2],
                                         cls.release, [cls.milestone])
        # add an update that fixes a bug which is an accepted Beta FE
        # and a proposed Beta blocker, and another bug which is an
        # accepted Final blocker, for fix function tests
        betafebug = add_bug(9003, 'betafebug', cls.milestone)
        betafebug.accepted_blocker = False
        betafebug.proposed_blocker = True
        betafebug.accepted_fe = True
        cls.update_complex = add_update(u'test-complex1.fc99', u'testing',
                                        [betafebug, finalbug],
                                        cls.release, [cls.milestone, cls.finalmile])
        # also an update that fixes a proposed Beta FE and an accepted
        # Final FE for the same purpose
        propbetafebug = add_bug(9004, 'propbetafebug', cls.milestone)
        propbetafebug.accepted_blocker = False
        propbetafebug.proposed_fe = True
        finalfebug = add_bug(9005, 'finalfebug', cls.finalmile)
        finalfebug.accepted_blocker = False
        finalfebug.accepted_fe = True
        cls.update_complexfe = add_update(u'test-complexfe.fc99', u'testing',
                                          [propbetafebug, finalfebug],
                                          cls.release, [cls.milestone, cls.finalmile])
        db.session.commit()

    @classmethod
    def teardown_class(cls):
        db.session.rollback()
        db.drop_all()

    def test_get_pending_stable_updates(self):
        updates = main.get_milestone_pending_stable_updates(self.milestone)
        assert len(updates) == 1
        update = updates[0]
        assert self.update_pending_stable == update


    def test_get_testing_status_and_has_bugs_updates(self):
        updates = main.get_milestone_updates_testing(self.milestone)
        assert len(updates) == 3
        expected = [self.update_testing1, self.update_complex, self.update_complexfe]
        updates.sort(key=lambda x: x.title)
        expected.sort(key=lambda x: x.title)
        assert updates == expected

    def test_get_milestone_all_nonstable_blocker_fixes(self):
        updates = main.get_milestone_all_nonstable_blocker_fixes(self.milestone)
        assert len(updates) == 3
        # we should NOT find complex or complexfe here!
        expected = [self.update_pending_stable, self.update_pending_testing, self.update_testing1]
        updates.sort(key=lambda x: x.title)
        expected.sort(key=lambda x: x.title)
        assert updates == expected

    def test_get_milestone_all_nonstable_fe_fixes(self):
        updates = main.get_milestone_all_nonstable_fe_fixes(self.milestone)
        # we should find complex (as it fixes an accepted Beta FE) but
        # NOT complexfe here!
        assert len(updates) == 1
        update = updates[0]
        assert self.update_complex == update

    def test_requests(self):
        # we also test the requests template generation here, as it's
        # what the _fixes queries back and it makes sense to do it
        # with all these bits in place
        with app.app_context():
            beta_response = main.display_release_requests(num=self.release.number,
                                                          release_name=self.milestone.version)
            final_response = main.display_release_requests(num=self.release.number,
                                                           release_name=self.finalmile.version)
        # === beta milestone ===
        beta_requests = beta_response.get_data(as_text=True)
        print(beta_requests)
        # split up the text
        all_sections = split_requests(beta_requests)
        compose_blockers, compose_fes, push_blockers, push_fes, deps = all_sections

        # check all the right updates and bugs are and are not in the correct sections
        assert item_just_in("test-pending-stable", [compose_blockers, push_blockers], all_sections)
        assert item_just_in("test-stable1", [], all_sections)
        assert item_just_in("test-stable2", [], all_sections)
        assert item_just_in("test-pending-testing", [compose_blockers], all_sections)
        assert item_just_in("test-testing1", [compose_blockers], all_sections)
        assert item_just_in("test-testing2", [], all_sections)
        assert item_just_in("test-complex1", [compose_fes], all_sections)
        assert item_just_in("test-complexfe", [], all_sections)
        # bug 1 id
        assert item_just_in("9000", [compose_blockers, push_blockers, deps], all_sections)
        # bug 2 id
        assert item_just_in("9001", [], all_sections)
        # finalbug id
        assert item_just_in("9002", [], all_sections)
        # betafebug id
        assert item_just_in("9003", [compose_fes], all_sections)
        # propbetafebug id
        assert item_just_in("9004", [], all_sections)
        # finalfebug id
        assert item_just_in("9005", [], all_sections)

        # === final milestone ===
        final_requests = final_response.get_data(as_text=True)
        print(final_requests)
        # split up the text
        all_sections = split_requests(final_requests)
        compose_blockers, compose_fes, push_blockers, push_fes, deps = all_sections

        # check all the right updates and bugs are and are not in the correct sections
        assert item_just_in("test-pending-stable", [], all_sections)
        assert item_just_in("test-stable1", [], all_sections)
        assert item_just_in("test-stable2", [], all_sections)
        assert item_just_in("test-pending-testing", [], all_sections)
        assert item_just_in("test-testing1", [], all_sections)
        assert item_just_in("test-testing2", [], all_sections)
        assert item_just_in("test-complex1", [compose_blockers], all_sections)
        assert item_just_in("test-complexfe", [compose_fes], all_sections)
        # bug 1 id
        assert item_just_in("9000", [], all_sections)
        # bug 2 id
        assert item_just_in("9001", [], all_sections)
        # finalbug id
        assert item_just_in("9002", [compose_blockers], all_sections)
        # betafebug id
        assert item_just_in("9003", [], all_sections)
        # propbetafebug id
        assert item_just_in("9004", [], all_sections)
        # finalfebug id
        assert item_just_in("9005", [compose_fes], all_sections)
